from sqlalchemy import Column, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

class SQLTestRecord(Base):
    __tablename__ = 'test-records'
    test_uuid = Column(String, primary_key=True)
    data = Column(JSONB)



import json
import jwt
import time

from functools import lru_cache
from dataclasses import dataclass

from json.decoder import JSONDecodeError
from simplejson.scanner import JSONDecodeError as SimpleJSONDecodeError

from simple_memory_cache import GLOBAL_CACHE

from ..errors import AuthUnauthorized, ExpiredAccessToken, SpintopException

from spintop.utils import spintop_session

class Auth0Error(AuthUnauthorized):
    def __init__(self, resp):
        self.content = _get_resp_content(resp)
        self.resp = resp
        super().__init__(self._create_error_message())

    def _create_error_message(self):
        desc = self.error_description
        if not desc:
            desc = str(self.content)
        return f'Request {_request_to_string(self.resp.request)!r} returned an error response: ({self.resp.status_code}) {desc}'

    @property
    def error(self):
        return self.content.get('error', None)

    @property
    def error_description(self):
        return self.content.get('error_description', None)

def _get_resp_content(resp):
    try:
        content = resp.json()
    except (JSONDecodeError, SimpleJSONDecodeError):
        content = {}
    return content

def _request_to_string(request):
    return f'{request.method} {request.url}'

class BaseAuth0Provider(object):
    def __init__(self, domain, audience, client_id, jwks_url, user_info_url, secret_key=None):
        self.domain = domain
        self.audience = audience
        self.client_id = client_id
        self.jwks_url = jwks_url
        self.user_info_url = user_info_url
        self.secret_key = secret_key

        if None in [domain, audience, client_id, jwks_url, user_info_url]:
            raise ValueError('Auth0Provider requires its parameters to not be None.')

        self.cached_jwks = GLOBAL_CACHE.MemoryCachedVar('auth_jwks')
        self.cached_jwks.on_first_access(self._retrieve_jwks)

    @property
    @lru_cache()
    def session(self):
        return spintop_session()

    def _retrieve_jwks(self):
        return self.session.get(self.jwks_url).json()
    
    def get_user_info(self, access_token):
        data = self.session.get(self.user_info_url, headers=self.get_auth_headers(access_token))
        return data.json()

    def get_auth_headers(self, access_token):
        return {'Authorization': 'Bearer ' + access_token}

    def _execute_request(self, url_path, **data):
        URL = 'https://' + self.domain  + url_path
        resp = self.session.post(URL, data = data)
        content = self._handle_errors_in_content(resp)
        return content
    
    def _handle_errors_in_content(self, resp):
        if not resp.ok:
            raise Auth0Error(resp)
    
        return _get_resp_content(resp)

    def get_jwks_key(self, access_token):
        jwks = self.cached_jwks.get()
        public_keys = {}
        for jwk in jwks['keys']:
            kid = jwk['kid']
            public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
            
        kid = jwt.get_unverified_header(access_token)['kid']
        key = public_keys[kid]
        return key

    def jwt_verify(self, access_token):
        try:
            payload = self.decode(access_token)
        except jwt.ExpiredSignatureError:
            raise ExpiredAccessToken("Token Expired")
        except jwt.PyJWTError:
            # invalidate jwks cache and try ONCE more.
            self.cached_jwks.invalidate()
            payload = self.decode(access_token)
        return payload
    
    def decode(self, access_token):
        key = self.get_jwks_key(access_token)
        return jwt.decode(access_token, key=key, algorithms=['RS256'], audience=[self.client_id, self.audience])

class Auth0Provider(BaseAuth0Provider):
    def client_credentials(self):
        if not self.secret_key:
            raise SpintopException('Client credentials grant requires a secret key.')

        return self._execute_request('/oauth/token',
            grant_type = 'client_credentials',
            client_id = self.client_id,
            client_secret = self.secret_key,
            audience = self.audience,
        )

    def authenticate(self, username, password, scopes=[]):
        return self._execute_request('/oauth/token',
            grant_type = 'password',
            username = username,
            password = password,
            audience = self.audience,
            client_id = self.client_id,
            scope = ' '.join(scopes)
        )

    def DeviceAuthorizationFlow(self):
        return DeviceAuthorizationFlow(self)

    def request_device_code(self, scopes=[]):
        return self._execute_request('/oauth/device/code',
            audience = self.audience,
            client_id = self.client_id,
            scope = ' '.join(scopes)
        )

    def check_device_code_approval(self, device_code):
        return self._execute_request('/oauth/token',
            grant_type = 'urn:ietf:params:oauth:grant-type:device_code',
            device_code = device_code,
            client_id = self.client_id
        )
    
    def refresh_access_token(self, refresh_token):
        if not refresh_token:
            raise AuthUnauthorized('No refresh token exists; unable to execute operation.')

        return self._execute_request('/oauth/token',
            grant_type = 'refresh_token',
            client_id = self.client_id,
            refresh_token = refresh_token
        )

    def revoke_refresh_token(self, refresh_token):
        self._execute_request('/oauth/revoke',
            client_id = self.client_id,
            token = refresh_token
        )
    

class DeviceAuthorizationFlow(object):
    def __init__(self, auth0_provider):
        self.auth0 = auth0_provider
        self.device_code_response = None

    def request_verification_url(self, scopes=[]):
        raw_response = self.auth0.request_device_code(scopes=scopes)
        self.device_code_response = DeviceCodeResponse(**raw_response)
        return self.device_code_response

    def wait_for_approval(self):
        interval = self.device_code_response.interval
        expires_in = self.device_code_response.expires_in
        device_code = self.device_code_response.device_code

        start = time.time()

        response = None

        while time.time() - start < expires_in:
            try:
                response = self.auth0.check_device_code_approval(device_code)
                break
            except Auth0Error as e:
                if e.error == "expired_token":
                    raise AuthUnauthorized("Did not receive an access token in time.")
                elif e.error == "access_denied":
                    raise AuthUnauthorized("User did not accept access.")
                elif e.error == "slow_down":
                    interval = interval + 2
                else:
                    # authorization_pending
                    pass
                    
                time.sleep(interval)
        
        if not response:
            raise AuthUnauthorized("No access token received.")
            
        return response
    
@dataclass
class DeviceCodeResponse():
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int



import inspect
from reprlib import recursive_repr
from datetime import datetime
from collections.abc import Mapping

from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field, asdict

from spintop.utils import GetRecursiveMixin, GetRecursive

from .view import DataClassPrimitiveView, join_key_prefix
from .meta import MetaBaseDataClass, MetaField

_TYPE_MAP = {}

def register_type(_type):
    def _wrap_cls(cls):
        cls._type = _type
        assert _type not in _TYPE_MAP, "Type %s is already registered to class %s" % (_type, _TYPE_MAP[_type])
        _TYPE_MAP[_type] = cls
        return cls
    return _wrap_cls

def cls_of_obj(obj):
    try:
        _type = type_of(obj)
        cls = cls_of_type(_type)
        if cls is None:
            raise AttributeError()
        return cls
    except AttributeError:
        return cls_of_serialized(obj)

def cls_of_serialized(obj_serialized):
    return cls_of_type(serialized_type_of(obj_serialized))

def cls_of_type(_type):
    return _TYPE_MAP.get(_type, None)

def type_of(cls):
    return cls._type

def serialized_type_of(obj_serialized):
    return obj_serialized.get('_type', None)

def type_dict_of(cls):
    return {'_type': type_of(cls)}

def is_type_of(cls, base_cls):
    return type_of(cls) == type_of(base_cls)

def is_serialized_type_of(obj_serialized, base_cls):
    return serialized_type_of(obj_serialized) == type_of(base_cls)

class CacheByCls():
    
    def __init__(self):
        self.value_by_cls = {}
        
    def retrieve(self, cls):
        if not self.value_by_cls.get(cls, None):
            self.value_by_cls[cls] = self._get(cls)
        return self.value_by_cls[cls]
    
    def _get(self, cls):
        raise NotImplementedError()
    
class FieldsCache(CacheByCls):
    """The fields() function is quite slow. This allows to store its 
    return value per cls, as it shouldn't change dynamically during
    runtime."""
    
    def _get(self, cls):
        return fields(cls)
    
class DefaultViewsCache(CacheByCls):
    def _get(self, cls):
        return DataClassPrimitiveView(
            cls.get_default_view_mapping()
        )

_fields_cache = FieldsCache()
_default_views_cache = DefaultViewsCache()

def model_dataclass(**kwargs):
    """Custom dataclass decorator that removes the default values set on classes.
    This allows the meta base data class to return the field itself instead of the
    default value."""
    def _decorator(cls):
        cls = dataclass(init=False, repr=False, **kwargs)(cls)
        fields = cls.get_fields()
        for field in fields:
            try:
                delattr(cls, field.name)
            except AttributeError:
                pass
        return cls
    return _decorator

@model_dataclass()
@register_type('_base')
class BaseDataClass(GetRecursiveMixin, metaclass=MetaBaseDataClass):
    
    def __init__(self, *args, **kwargs):
        if args:
            raise NotImplementedError('Positionnal args not supported with BaseDataClass descendants.')
        defaults = self.get_defaults()

        self.update_attrs(defaults) # First set defaults
        self.update_attrs(kwargs) # Then constructor arguments.
        self.__post_init__()

    def __post_init__(self):
        pass

    @classmethod
    def defaults(cls, **others):
        return others

    @classmethod
    def get_defaults(cls, **other_fields):
        defaults = cls.defaults()
        defaults.update({
            field.name: 
                default_field_value(field) for field in cls.get_fields() if field.name not in defaults
        })
        return defaults
    
    @classmethod
    def null(cls, **not_null_fields):
        """ Creates an instance with all attributes set to None. 
        Calls null() on any sub- BaseDataClass fields.
        """
        arguments = {}

        for key, default in cls.get_defaults().items():    
            arguments[key] = default
        
        arguments.update(not_null_fields)
        return cls(**arguments)
    
    @classmethod
    def get_fields(cls):
        return _fields_cache.retrieve(cls)
    
    @classmethod
    def get_data_fields(cls):
        """ Sub classed by FeatureRecord below.
        """
        return cls.get_fields()
        
    @classmethod
    def get_default_view(cls):
        return _default_views_cache.retrieve(cls)
        
    @classmethod
    def get_default_view_mapping(cls, with_key_prefix=None):
        if with_key_prefix is not None:
            key_creator = lambda key: join_key_prefix(with_key_prefix, key)
        else:
            key_creator = lambda key: key
        
        # The default view uses the fields defined by the get_data_fields class method.
        fields = cls.get_data_fields()
        
        return {key_creator(field.name): MetaField(cls, field) for field in fields}
        
    def update_attrs(self, attributes_dict):
        for key, value in attributes_dict.items():
            setattr(self, key, value)
    
    def copy(self, with_data=True, as_cls=None, set_attributes={}):
        if with_data:
            nullify_fields = []
        else:
            nullify_fields = self.get_data_fields()
            
        copy_obj = copy_nullify_fields(self, nullify_fields, target_cls=as_cls)
        for name, value in set_attributes.items():
            setattr(copy_obj, name, value)
        return copy_obj
    
    def as_dict(self):
        """ Transforms this object recursively into a dict. Adds the static
        _type field to the data"""
        data = asdict(self)
        data.update(type_dict_of(self))
        return data
    
    @staticmethod
    def cls_data_from_dict(cls, data):
        _type = data.pop('_type', None)
        if _type:
            cls = cls_of_type(_type)
        return cls, data

    def getattr_from_field(self, meta_field):
        return getattr_from_field(self, meta_field)

    def hasattr_from_field(self, meta_field):
        return hasattr_from_field(self, meta_field)

    @recursive_repr()
    def __repr__(self):
        fields = tuple("{}={}".format(field.name, repr(getattr(self, field.name, None))) for field in self.get_fields())
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))

def getattr_from_field(obj, meta_field):
    obj_cls = cls_of_obj(obj)
    if isinstance(meta_field, property):
        return meta_field.__get__(obj, type(obj))
    elif meta_field.is_container_of_cls(obj_cls):
        return GetRecursive(obj).get_recursive(as_array=meta_field.name_tuple_)
    else:
        raise AttributeError('This class {!r} is not a subclass of the meta field type {!r}'.format(obj_cls, meta_field.container_type_))

def hasattr_from_field(obj, meta_field):
    try:
        getattr_from_field(obj, meta_field)
        return True
    except AttributeError:
        return False


def copy_nullify_fields(obj, null_fields, target_cls=None):
    if target_cls is None:
        target_cls = obj.__class__
    
    copy_obj = target_cls.null()
    not_null_fields = (set(obj.get_fields()) - set(null_fields)) & set(target_cls.get_fields())
    for field in not_null_fields:
        setattr(copy_obj, field.name, getattr(obj, field.name))
    
    return copy_obj

def default_field_value(field):
    if inspect.isclass(field.type) and issubclass(field.type, BaseDataClass):
        value = field.type.null()
    elif not isinstance(field.default_factory, _MISSING_TYPE):
        value = field.default_factory()
    elif not isinstance(field.default, _MISSING_TYPE):
        if callable(field.default):
            value = field.default()
        else:
            value = field.default
    else:
        value = None
    
    return value



import re
from collections.abc import Sequence

class MongoQuery():
    def __init__(self, model_query_or_queries):
        if not isinstance(model_query_or_queries, Sequence):
            model_query_or_queries = [model_query_or_queries]

        self.orig_queries = model_query_or_queries

    def build(self):
        if len(self.orig_queries) == 1:
            return self.build_one_query(self.orig_queries[0])
        else:
            return {"$or": [self.build_one_query(q) for q in self.orig_queries]}

    def build_one_query(self, orig_query):
        query = {}

        if orig_query is None:
            return query

        for key, value in orig_query.value_equals.items():
            query[key] = value
        
        for key, value in orig_query.list_contains.items():
            raise NotImplementedError()

        for key, list_of_values in orig_query.value_equals_one_of.items():
            query[key] = q_any_in_list(list_of_values)
        
        return query
    
def q_any_in_list(_list):
    return {'$in': _list}


from functools import lru_cache, wraps

import marshmallow_dataclass

@lru_cache()
def get_schema(message_model_cls):
    Schema = marshmallow_dataclass.class_schema(message_model_cls)
    return Schema()


from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any, Union

from spintop.models import SpintopTestRecord, SpintopSerializedFlatTestRecord

@dataclass
class SpintopEnvMessage():
    env: Optional[Dict[str, Optional[Any]]]

    def create_message_env(self, local_env):
        """Create a copy of the local env with the attributes contained in this message"""
        env = self.env
        if env is None:
            env = {}
        return local_env.copy(**env)

    @classmethod
    def from_message(cls, other_env_message, **kwargs):
        obj = cls(env=other_env_message.env, **kwargs)
        return obj

@dataclass
class TestRecordMessage(SpintopEnvMessage):
    test_record: Union[SpintopTestRecord, SpintopSerializedFlatTestRecord]

@dataclass
class TestsUpdateMessage(SpintopEnvMessage):
    updated_test_uuids: List[str]
    deleted_test_uuids: List[str]
    affected_duts: List[str]
    update_all: bool = False
    delete_all: bool = False

    @classmethod
    def create(self, **kwargs):
        defaults = dict(
            updated_test_uuids = [],
            deleted_test_uuids = [],
            affected_duts = [],
            update_all = False,
            delete_all = False,
            env = None
        )
        defaults.update(kwargs)
        return TestsUpdateMessage(**defaults)

@dataclass
class TestsUpdateQueryMessage(SpintopEnvMessage):
    query_dict: Dict[str, Any]



from itertools import chain

from collections import OrderedDict, defaultdict
from collections.abc import Mapping
from typing import NamedTuple, List

from spintop.utils import GetRecursive, GetRecursiveMixin, dict_ops

from ..base import (
    is_serialized_type_of, 
    type_dict_of
)
from ..view import DataClassPrimitiveView, ComplexPrimitiveView

from ._base import (
    BaseDataClass, 
    TestIDRecord, 
    TestRecordSummary, 
    FeatureRecord, 
    DefaultPrimitiveView, 
    compute_stats,
    serialized_get_test_uuid
)


SUPPORT_MISSING_TOP_LEVEL_RECORD = False

def normalize_features_len(dict_array):
    new_dict = {}
    max_column_key_len = max([len(key) for key in dict_array])
    for key, value in dict_array.items():
        if len(key) < max_column_key_len:
            key = key + ('',)*(max_column_key_len - len(key))
        new_dict[key] = value
    return new_dict
    
            
class SpintopTestRecordCollection(object):
    def __init__(self, flat_records=None):
        self.records = list(flat_records) if flat_records else []
    
    @property
    def collector(self):
        return self.add_record

    @property
    def test_uuids(self):
        return [record.test_uuid for record in self.records]

    def add_record(self, record):
        self.records.append(record)
        
    def apply(self, *fns):
        for record in self.iter_records():
            for fn in fns:
                fn(record)
                
    def iter_records(self):
        for record in self.records:
            yield record
                
    def count_unique(self, key=lambda record: record.test_id.test_uuid):
        occurances = set()
        self.apply(lambda record: occurances.add(key(record)))
        return len(occurances)
            
    def are_records_unique(self):
        test_uuids = self.test_uuids
        return len(test_uuids) == len(set(test_uuids))

    def sort(self, key=lambda record: record.test_id.start_datetime):
        self.records.sort(key=key)
        
    def avg_feature_count(self):
        feature_counts = [len(record.features) for record in self.iter_records()]
        return sum(feature_counts)/len(feature_counts)
    
    def __eq__(self, other):
        return self.records == other.records

class SpintopTestRecord(GetRecursiveMixin):
    def __init__(self, test_record, features=None):
        self._data = None
        self._features = None
        
        if test_record is None:
            raise ValueError('None test_record is not supported.')

        self.data = test_record # The top-level data
        self.features = tuple(features) if features else tuple()
    
    @property
    def features(self):
        return self._features if self._features else tuple()
    
    @features.setter
    def features(self, features):

        self._broadcast_test_id(features)
        self._features = tuple(sorted(features, key=lambda feature: feature.index))
        
    @property
    def all_features(self):
        return (self.data,) + self.features

    def compute_stats(self):
        if not self.data.feature_count:
            compute_stats(self.all_features)

    @property
    def data(self):
        return self._data

    @property
    def test_record(self):
        return self.data
    
    @data.setter
    def data(self, value):
        self._data = value
        self._broadcast_test_id(self.features)
        
    @property
    def test_id(self):
        return self.data.test_id

    @property
    def test_uuid(self):
        return self.test_id.test_uuid
    
    def __hash__(self):
        return hash(self.test_uuid)

    def __eq__(self, other):
        return self.data == other.data and self.features == other.features
    
    def __repr__(self):
        return '{}(test_record={!r}, features=[...]*{})'.format(self.__class__.__name__, self.data, len(self.features))

    def remove_duplicate_features(self):
        feature_names = set()
        features = []
        for feature in self.features:
            if feature.name in feature_names:
                continue
            else:
                features.append(feature)
                feature_names.add(feature.name)
        
        self.features = features
    
    def _broadcast_test_id(self, features):
        """Uses the top level test_id and sets it for all features, ensuring
        that all features have the same TestIDRecord reference."""
        for feature in features:
            feature.test_id = self.data.test_id
    
    def reindex(self):
        for index, feature in enumerate(self.features):
            feature.index = index

    def normalize_outcomes(self):
        parents = tuple()
        last_feature = None
        for feature in self.all_features:
            while feature.depth > len(parents):
                if last_feature.name not in feature.ancestors:
                    raise Exception('Unable to normalize outcomes: feature order does not match ancestors. {!r} should be in ancestors: {!r}'.format(
                        last_feature.name,
                        feature.ancestors
                    )) 
                parents = parents + (last_feature,)

            if feature.depth < len(parents):
                parents = parents[:feature.depth]

            outcome = feature.outcome

            for parent in reversed(parents):
                outcome.impose_upon(parent.outcome)
                outcome = parent.outcome

            last_feature = feature

            
    def fill_missing_from_source(self, fill_source, on_fill=None):
        max_current_feature_index = max(self.features, key=lambda f: f.index)
        new_features_len = len(fill_source)
        
        assert new_features_len >= max_current_feature_index.index, "Fill source must be same length or bigger than this test_record max index."
        
        new_features = [None]*new_features_len
        
        if on_fill is None:
            on_fill = lambda obj: obj.copy()
            
        def fill_range(start, end):
            for i in range(start, end):
                new_features[i] = on_fill(fill_source[i])
        
        current_index = 0
        for feature in sorted(self.features, key=lambda f: f.index):
            fill_range(current_index, feature.index)
            new_features[feature.index] = feature # Keep this feature
            current_index = feature.index + 1
        
        fill_range(current_index, new_features_len) # Fill end
        
        self.features = new_features
                
    def find_feature(self, condition, start_index=0):
        try:
            return next(feat for feat in self.all_features[start_index:] if condition(feat))
        except StopIteration:
            raise ValueError('No feature matched condition.')
    
    def add_tag(self, key, value=True):
        self.test_id.add_tag(key, value=value)
        
    def remove_tag(self, key):
        self.test_id.remove_tag(key)

    def as_dict(self):
        return dict(
            test_record=self._data,
            features=list(self._features)
        )

    @classmethod
    def from_dict(cls, _dict):
        return cls(**_dict)

class SpintopTestRecordView(object):
    feature_prefix = ('features', lambda feat: feat._type, lambda feat: feat.complete_name)
    data_prefix = () 
    default_view = DefaultPrimitiveView()
    
    def __init__(self, feature_mapping=None, include_test_id=True):
        self.include_test_id = include_test_id

        if feature_mapping:
            self.feature_view = DataClassPrimitiveView(feature_mapping)
        else:
            self.feature_view = self.default_view

    def apply(self, record, flatten_dict=True):
        data = OrderedDict()
        
        if self.include_test_id:
            test_id_data = self.apply_default_test_id(record, flatten_dict=flatten_dict)
            dict_ops.update(data, test_id_data)
        
        for feature in record.features:
            feature_data = self.feature_view.apply(feature, flatten_dict=flatten_dict)
            dict_ops.update(data, feature_data)
        
        data = dict_ops.replace_defaults(data)

        return data
    
    def apply_default_feature(self, feature, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.feature_prefix
        return self.feature_view.apply(feature, key_prefix=key_prefix, **apply_kwargs)
    
    def apply_default_test_id(self, record, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.data_prefix
        data = self.default_view.apply(record.data, key_prefix=key_prefix, **apply_kwargs)
        return data

class SpintopFlatTestRecordBuilder(NamedTuple):
    test_record: TestRecordSummary
    features: List[FeatureRecord]
    
    def build(self):
        record = SpintopTestRecord(test_record=self.test_record, features=sorted(self.features, key=lambda feature: feature.index))
        record.compute_stats()
        return record

class SpintopSerializedFlatTestRecord(GetRecursiveMixin):

    def __init__(self, test_record=None, features=[]):
        self.test_record = test_record
        self.features = features

    @property
    def data(self):
        return GetRecursive(self.test_record)

    @property
    def test_id(self):
        return GetRecursive(self.test_record['test_id'])

    @property
    def test_uuid(self):
        return self.test_id['test_uuid']

    @property
    def all_features(self):
        return (self.test_record,) + tuple(self.features)

    def as_dict(self):
        return dict(
            test_record=self.test_record,
            features=self.features
        )

    def deserialize(self, serializer):
        builder = SpintopFlatTestRecordBuilder(
            test_record=serializer.deserialize(BaseDataClass, self.test_record), 
            features=tuple(serializer.deserialize(BaseDataClass, feat) for feat in self.features)
        )
        return builder.build()

    def __repr__(self):
        return '{}(test_record={!r}, features=[...]*{})'.format(self.__class__.__name__, self.test_record, len(self.features))

    def __eq__(self, other):
        return self.as_dict() == other.as_dict()

class SpintopSerializedTestRecordCollection(object):

    def __init__(self, records=[]):
        self.records = [
            SpintopSerializedFlatTestRecord(**record) if isinstance(record, Mapping) else record 
            for record in records
        ]

    @classmethod
    def from_features(cls, features):
        return cls(records=serialized_features_to_record_list(features))

    def deserialize(self, serializer):
        return (record.deserialize(serializer) for record in self.records)

def serialized_features_to_record_list(features):
    test_uuid_record_map = OrderedDict()
    
    for feature in features:
        found_test_record = None
        found_feature = None

        if is_serialized_type_of(feature, TestRecordSummary):
            found_test_record = feature
        else:
            found_feature = feature
        
        test_uuid = serialized_get_test_uuid(feature)

        record = test_uuid_record_map.get(test_uuid, None)
        if record is None:
            record = SpintopSerializedFlatTestRecord(test_record=None, features=[])
        
        if found_test_record:
            if record.test_record is not None:
                raise ValueError('Two features match as a top level record for the same record: {!r} and {!r}'.format(record.test_record, found_test_record))
            record.test_record = found_test_record
        else:
            record.features.append(found_feature)

        test_uuid_record_map[test_uuid] = record
    
    records = list(test_uuid_record_map.values())

    for record in records:
        if record.test_record is None:
            # Find a valid test_id from features.
            if SUPPORT_MISSING_TOP_LEVEL_RECORD:
                test_id = None
                if record.features:
                    test_id = record.features[0]['test_id']
                else:
                    raise ValueError('Empty Test Record.')
                record.test_record = TestRecordSummary.null(test_id=test_id, original=False).as_dict()
            else:
                raise RuntimeError('Missing top level record from SpintopTestRecord {!r}'.format(record))

    return records

        
        
    


from collections.abc import Mapping

class GetRecursiveMixin(object):
    
    def get_recursive(self, as_str=None, as_array=None):
        if as_str:
            as_array = as_str.split('.')
        
        this_attr_name, *as_array = as_array
        this_attr = self._get_attr_recursive(this_attr_name)
        

        if as_array:
            if not isinstance(this_attr, GetRecursiveMixin):
                this_attr = AnonymousGetRecursive(this_attr)
            return this_attr.get_recursive(as_array=as_array)
        else:
            return this_attr
    
    def _get_attr_recursive(self, name):
        return getattr(self, name)

class GetRecursiveKeyMixin(GetRecursiveMixin):

    def _get_attr_recursive(self, name):
        return self[name]

class AnonymousGetRecursive(Mapping, GetRecursiveMixin):
    def __init__(self, obj):
        self.obj = obj

    def _get_attr_recursive(self, name):
        if isinstance(self.obj, Mapping):
            return self.obj[name]
        else:
            try:
                return getattr(self.obj, name)
            except AttributeError:
                return self.obj[name]

    def __iter__(self):
        return self.obj.__iter__()

    def __len__(self):
        return self.obj.__len__()

    def __getattr__(self, name):
        return self._get_attr_recursive(name)

    def __getitem__(self, key):
        return self._get_attr_recursive(key)

GetRecursive = AnonymousGetRecursive



from functools import wraps
from collections import defaultdict

from .models import TestsUpdateMessage
from .models import TestsUpdateQueryMessage

from .schemas import get_schema

TESTS_UPDATE_TOPIC = 'tests.update'
TESTS_UPDATE_QUERY_TOPIC = 'tests.update_query'

def messages_endpoint(topic_name, message_model_cls):
    return property(
        fget = lambda publisher: _MessagesEndpoint(publisher, topic_name, message_model_cls)
    )

class SpintopMessagePublisher(object):
    tests_update = messages_endpoint(TESTS_UPDATE_TOPIC, TestsUpdateMessage)
    tests_query_update = messages_endpoint(TESTS_UPDATE_QUERY_TOPIC, TestsUpdateQueryMessage)

    def subscribe(self, topic, callback):
        raise NotImplementedError()

    def publish(self, topic, message):
        raise NotImplementedError()

    def prepare_message(self, message):
        pass

    def supports_publish(self):
        return True

    def supports_subscribe(self):
        return True

class _MessagesEndpoint(object):
    def __init__(self, publisher, topic, message_model_cls):
        self.publisher = publisher
        self.topic = topic
        self.message_model_cls = message_model_cls
    
    @property
    def schema(self):
        return get_schema(self.message_model_cls)

    def subscribe(self, callback):
        if self.publisher.supports_subscribe():
            @wraps(callback)
            def _wrapper(raw_message):
                message = self.load_message(raw_message)
                callback(message)

            self.publisher.subscribe(self.topic, _wrapper)

    def dump_message(self, message):
        return self.schema.dump(message)

    def load_message(self, raw_message):
        return self.schema.load(raw_message)

    def publish(self, message):
        if self.publisher.supports_publish():
            self.publisher.prepare_message(message)
            content = self.dump_message(message)
            self.publisher.publish(self.topic, content)

class LocalMessagePublisher(SpintopMessagePublisher):
    def __init__(self, local_env):
        self._local_subscribers = defaultdict(list)
        self._env = local_env

    def subscribe(self, topic, callback):
        self._local_subscribers[topic].append(callback)

    def publish(self, topic, message):
        self._publish(topic, message)
        self._notify_local(topic, message)

    def prepare_message(self, message):
        message.env = self._env.freeze_database_access_only()
        super().prepare_message(message)

    def _notify_local(self, topic, message):
        for subscriber in self._local_subscribers[topic]:
            subscriber(message)

    def _publish(self, topic, message):
        pass

class NoopMessagePublisher(SpintopMessagePublisher):
    
    def supports_publish(self):
        return False

    def supports_subscribe(self):
        return False


import os
from incremental_module_loader import IncrementalModuleLoader

from .auth import AuthModule, FilePathCredentialsStore
from .logs import _logger, setup_logging
from .storage import SITE_DATA_DIR
from .api_client import SpintopAPIClientModule, SpintopAPISpecAuthBootstrap

logger = _logger('root')

def SpintopFactory(
        api_url,
        credentials_filepath=None,
        selected_org_id=None,
        logs_factory=setup_logging,
        auth_bootstrap_factory=SpintopAPISpecAuthBootstrap,
        credentials_store_factory=FilePathCredentialsStore,
        auth_factory=AuthModule, 
        spintop_api_factory=SpintopAPIClientModule,
        final_factory=None
    ):

    loader = IncrementalModuleLoader()
    loader.update(
        api_url=api_url,
        credentials_filepath=credentials_filepath,
        selected_org_id=selected_org_id
    )
    
    loader.load(logs=logs_factory)
    loader.load(credentials_store=credentials_store_factory)
    loader.load(auth_bootstrap=auth_bootstrap_factory)
    loader.load(auth=auth_factory)
    spintop_or_final = loader.load(spintop_api=spintop_api_factory)
    

    if final_factory:
        spintop_or_final = loader.load(final_factory)
    
    return spintop_or_final

def SpintopWorkerFactory(worker_cls, **factory_kwargs):
    worker = SpintopFactory(final_factory=worker_cls, **factory_kwargs)
    return worker
        
        
    
    
        


from ._collection import (
    SpintopTestRecordCollection, 
    SpintopFlatTestRecordBuilder,
    SpintopSerializedFlatTestRecord,
    SpintopTestRecord,
    SpintopTestRecordView,
    SpintopSerializedTestRecordCollection
)

from ._base import (
    serialized_get_test_uuid,
    DefaultPrimitiveView,
    BaseDataClass,
    TestIDRecord,
    TestRecordSummary, 
    FeatureRecord,
    MeasureFeatureRecord,
    PhaseFeatureRecord,
    DutIDRecord,
    DutOp,
    OutcomeData,
    NO_VERSION
)

from ._tree_struct import SpintopTreeTestRecord, SpintopTreeTestRecordBuilder


import os
from .schemas import credentials_schema

from spintop.utils import write_yaml_file, load_yaml_file

class CredentialsStore(object):
    def store(self, credentials):
        raise NotImplementedError()

    def retrieve(self):
        raise NotImplementedError()
    
    def delete(self):
        raise NotImplementedError()

class InMemoryCredentialsStore(object):
    def __init__(self):
        self.credentials = None

    def store(self, credentials):
        self.credentials = credentials

    def retrieve(self):
        return self.credentials
    
    def delete(self):
        self.credentials = None

class FilePathCredentialsStore(object):
    def __init__(self, credentials_filepath): # Do not change param name for module compat
        self.filepath = credentials_filepath

    def store(self, credentials):
        credentials = credentials_schema.dump(credentials)
        write_yaml_file(self.filepath, credentials)

    def retrieve(self):
        try:
            content = load_yaml_file(self.filepath)
        except FileNotFoundError:
            return None
        else:
            return credentials_schema.load(content)
    
    def delete(self):
        if os.path.exists(self.filepath):
            os.remove(self.filepath)



from bson.objectid import ObjectId

from spintop.models import BaseDataClass
from spintop.models.serialization import get_bson_serializer

def to_mongo_id(_dict, strip_id_none=False):
    if 'oid' in _dict:
        _dict['_id'] = to_object_id(_dict.pop('oid'))
    
    if strip_id_none and _dict.get('_id', False) is None:
        del _dict['_id']
        
    return _dict
    
def from_mongo_id(_dict):
    if '_id' in _dict:
        _dict['oid'] = from_object_id(_dict.pop('_id'))
    return _dict

def to_object_id(encoded_bytes):
    if encoded_bytes:
        return ObjectId(encoded_bytes)
    else:
        return None
    
def from_object_id(oid):
    if oid:
        return str(oid)
    else:
        return None

class DataClassSerializer(object):
    def __init__(self, serializer=None):
        if serializer is None: serializer = get_bson_serializer()
        
        self.serializer = serializer
        
    def serialize(self, obj):
        return self.serializer.serialize(obj)
    
    def serialize_many(self, objs):
        return [self.serialize(obj) for obj in objs]
    
    def deserialize(self, serialized):
        return self.serializer.deserialize(BaseDataClass, serialized)

    def deserialize_many(self, objs):
        return [self.deserialize(obj) for obj in objs]

    def serialized_to_mongo_id(self, serialized, keep_id_if_none=True):
        return to_mongo_id(serialized, strip_id_none=not keep_id_if_none)

    def serialized_from_mongo_id(self, serialized):
        return from_mongo_id(serialized)
    


from spintop.generators import Generator

from spintop.models import (
    SpintopSerializedFlatTestRecord,
    TestRecordSummary, 
    FeatureRecord, 
    SpintopTestRecord,
    Query
)

from spintop.utils import AnonymousGetRecursive


from ..logs import _logger
from ..errors import SpintopException, spintop_warn
from ..messages import NoopMessagePublisher, TestsUpdateMessage

logger = _logger('persistence')


class MissingMapper(SpintopException):
    def __init__(self, cls):
        super().__init__("Mapper for class {!r} is mandatory.".format(cls))

class NoMapperForObject(SpintopException):
    def __init__(self, obj, mappers):
        super(NoMapperForObject, self).__init__(
            'There are no known mapper able to interact with obj {!r} of class {}. Declared mappers are: {}'.format(
                obj,
                obj.__class__,
                [cls.__name__ for cls in mappers]
            )
        )
        
class DuplicateMapperClassName(SpintopException):
    def __init__(self, objcls):
        super(DuplicateMapperClassName, self).__init__(
            'The name of the class {} is duplicate. The class name linked to a mapper must be unique.'.format(
                objcls,
            )
        )

class TestRecordNotFound(SpintopException):
    def __init__(self, test_uuid):
        super().__init__(f'Test Record with UUID {test_uuid!r} not found.')

class ManyTestRecordFound(SpintopException):
    def __init__(self, test_uuid, count):
        super().__init__(f'{count} Test Records where found with UUID {test_uuid!r}, but only one was expected.')


class PersistenceFacade(object):
    logger = logger
    def __init__(self, serializer, messages=None):
        self.serializer = serializer

        if messages is None:
            # Will discard all messages
            messages = NoopMessagePublisher()

        self.messages = messages

    @classmethod
    def from_env(cls, uri, database_name, env=None):
        raise NotImplementedError()

    def _publish_test_record_update(self, test_records):
        modified_test_uuids = set(tr.test_uuid for tr in test_records)

        dut_id_field = TestRecordSummary.test_id.dut.id.name_

        modified_duts = set(AnonymousGetRecursive(tr.test_record).get_recursive(dut_id_field) for tr in test_records)

        self.messages.tests_update.publish(TestsUpdateMessage.create(
            updated_test_uuids=list(modified_test_uuids),
            affected_duts=list(modified_duts)
        ))

    def _publish_update_all(self):
        spintop_warn('Current update all deletes and re-creates all tests instead of targeting deleted uuids. This is very inefficient.')
        self.messages.tests_update.publish(TestsUpdateMessage.create(
            delete_all=True,
            update_all=True
        ))

    def serialize_barrier(self, records):
        records = list(records)
        if records and isinstance(records[0], SpintopTestRecord):
            records = [self.serializer.serialize(obj) for obj in records]

        if records and isinstance(records[0], dict):
            return [SpintopSerializedFlatTestRecord(**record) for record in records]
        else:
            return records

    def create(self, records):
        records = self.serialize_barrier(records)
        ret_val = self._create(records)
        self._publish_test_record_update(records)
        return ret_val

    def _create(self, records):
        raise NotImplementedError()
        
    def retrieve(self, query=None, deserialize=True, limit_range=None):
        if query is None: query = Query()
        serialized_collection = self._retrieve(query, limit_range=limit_range)
        if deserialize:
            yield from serialized_collection.deserialize(self.serializer)
        else:
            yield from (record for record in serialized_collection.records)

    def _retrieve(self, query, limit_range=None):
        """Should return a ``SpintopSerializedTestRecordCollection``"""
        raise NotImplementedError()
        
    def retrieve_one(self, test_uuid, deserialize=True):
        records = self.retrieve(Query().test_uuid_is(test_uuid), deserialize=deserialize)
        records = list(records)

        count = len(records)
        if count < 1:
            raise TestRecordNotFound(test_uuid)
        elif count > 1:
            raise ManyTestRecordFound(test_uuid, count)
        else:
            return records[0]
    
    def count(self, query=None):
        if query is None: query = Query()
        return self._count(query)
    
    def _count(self, query):
        raise NotImplementedError()

    def update(self, records, upsert=False):
        records = self.serialize_barrier(records)
        ret_val = self._update(records, upsert=upsert)
        self._publish_test_record_update(records)
        return ret_val

    def _update(self, records, upsert=False):
        raise NotImplementedError()
    
    def delete(self, query=None):
        if query is None: query = Query()
        ret_val = self._delete(query)
        self._publish_update_all()
        return ret_val

    def _delete(self, query=None):
        raise NotImplementedError()
    
    def create_records_generator(self):
        return PersistenceGenerator(self)
        
def create_mapper_name_index(mappers):
    mappers_name_index = {}
    for mapped_cls, mapper in mappers.items():
        name = mapped_cls.__name__
        if name in mappers_name_index:
            raise DuplicateMapperClassName(mapped_cls)
        mappers_name_index[name] = mapper
    return mappers_name_index
    
class PersistenceGenerator(Generator):
    def __init__(self, facade):
        super().__init__()
        self.facade = facade
    
    def __call__(self, *args, **kwargs):
        return self.facade.retrieve(*args, **kwargs)

class Mapper(object):
    def init(self):
        pass


from marshmallow import Schema, fields

class ManyTestsSchema(Schema):
    tests = fields.List(fields.Dict())
    
class TestsCountSchema(Schema):
    count = fields.Int()

tests_schema = ManyTestsSchema()
test_count_schema = TestsCountSchema()




import os

from spintop.persistence import PersistenceFacade
from spintop.models import (
    SpintopTestRecord, 
    SpintopSerializedFlatTestRecord, 
    SpintopSerializedTestRecordCollection, 
    TestRecordSummary,
    Query, 
    serialized_get_test_uuid,
    get_json_serializer
)

from spintop.errors import SpintopException

from .models import SQLTestRecord
from .operations import engine_from_uri, SQLOperations
from .queryset import PostgreSQLQuery

class PostgreSQLPersistenceFacade(PersistenceFacade):
    def __init__(self, sql_engine):
        self.sql_ops = SQLOperations(sql_engine)
        self.serializer = get_json_serializer()
        super().__init__(self.serializer)
        self.init()

    def init(self):
        self.sql_ops.create_tables()
    
    @classmethod
    def from_env(cls, uri, database_name, env=None):
        return cls(engine_from_uri(uri, database_name))
        
    def _create(self, records):
        records = self._to_sql_records(records)
        session = self.sql_ops.new_session()
        session.add_all(records)
        session.commit()
    
    def _to_sql_records(self, records):
        return [SQLTestRecord(test_uuid=record.test_uuid, data=record.as_dict()) for record in records]

    def count(self, query=None):
        if query is None: query = Query()
        session = self.sql_ops.new_session()
        operation = session.query(SQLTestRecord).filter(*self._to_filter(query))
        return operation.count()
    
    def avg_obj_size(self):
        raise NotImplementedError()
        
    def _retrieve(self, query=None, limit_range=None):
        """ Retrieve is responsible with returning the test_record (top level phase) associated
        with the matched features.
        """
        session = self.sql_ops.new_session()

        if query is None: query = Query()
        
        operation = session.query(SQLTestRecord).filter(*self._to_filter(query))
        if limit_range:
            operation = operation.offset(limit_range[0]).limit(limit_range[1] - limit_range[0])

        records = operation.all()
        records = [record.data for record in records]
        return SpintopSerializedTestRecordCollection(records)

    def _to_filter(self, query):
        return PostgreSQLQuery(query).build_data_query(SQLTestRecord)

    def _update(self, records, upsert=False):
        all_features = []
        test_uuids = []

        for record in records:
            all_features += record.all_features
            test_uuids += [record.test_uuid]

        session = self.sql_ops.new_session()

        session.query(SQLTestRecord).filter(*self._to_filter(Query().test_uuid_any_of(test_uuids))).delete(synchronize_session='fetch')
        records = self._to_sql_records(records)
        session.add_all(records)
        session.commit()
    
    def _delete(self, match_query):
        session = self.sql_ops.new_session()
        session.query(SQLTestRecord).filter(*self._to_filter(match_query)).delete(synchronize_session='fetch')
        session.commit()


import re
from sqlalchemy import or_, and_

from spintop.models import TestRecordSummary
from spintop.utils import AnonymousGetRecursive

class PostgreSQLQuery(object):
    def __init__(self, query):
        self.orig_query = query
    
    def build_data_query(self, test_record_table):
        """Only support test_uuids for now"""
        test_uuids = []
        queries = []

        for key, value in self.orig_query.value_equals.items():
            # data is stored as test_record: 
            if key == TestRecordSummary.test_id.test_uuid.name_:
                test_uuids.append(value)
            else:
                field = self._get_test_record_field(test_record_table, 'test_record.' + key)
                if isinstance(value, re.Pattern):
                    queries.append(field.astext.op("~")(value.pattern))
                else:
                    queries.append(field.astext == str(value))
        
        for key, value in self.orig_query.list_contains.items():
            raise NotImplementedError()

        for key, list_of_values in self.orig_query.value_equals_one_of.items():
            if key == TestRecordSummary.test_id.test_uuid.name_:
                test_uuids += list_of_values
            else:
                field = self._get_test_record_field(test_record_table, 'test_record.' + key)
                queries.append(field.in_(list_of_values))
        
        if test_uuids:
            queries.append(test_record_table.test_uuid.in_(test_uuids))

        return queries

    def _get_test_record_field(self, test_record_table, field_name):
        field = AnonymousGetRecursive(test_record_table.data)
        return field.get_recursive(field_name)


import warnings

class SpintopException(Exception):
    pass

class SpintopBaseException(BaseException):
    pass

class AuthUnauthorized(SpintopException):
    pass

class ExpiredAccessToken(SpintopException):
    pass

class SpintopWarning(UserWarning):
    pass

def spintop_warn(message):
    warnings.warn(message, SpintopWarning)


from datetime import datetime

from spintop.persistence.base import PersistenceFacade

from ..models import SpintopSerializedTestRecordCollection, Query, get_json_serializer, SpintopTestRecord

from .schemas import tests_schema, test_count_schema

class SpintopAPIPersistenceFacade(PersistenceFacade):
    def __init__(self, spintop_api):
        self.spintop_api = spintop_api
        self.serializer = get_json_serializer()
        super().__init__(self.serializer)

    @classmethod
    def from_env(self, uri, database_name=None, env=None):
        # database_name is the org_id
        api = env.spintop_factory()
        return api.tests

    @property
    def session(self):
        return self.spintop_api.session

    def _create(self, records):
        serialized = self._records_in_schema(records)
        return self.session.post(self.spintop_api.get_link('tests.create'), json=serialized)
        
    def _records_in_schema(self, records):
        return tests_schema.dump({'tests': [record.as_dict() for record in records]})

    def _retrieve(self, query, limit_range=None):
        query_dict = query.as_dict()
        if limit_range:
            query_dict['limit_range_inf'] = limit_range[0]
            query_dict['limit_range_sup'] = limit_range[1]

        resp = self.session.get(self.spintop_api.get_link('tests.retrieve'), params=query_dict)
        tests = tests_schema.load(resp.json())['tests']
        return SpintopSerializedTestRecordCollection(tests)

    # def retrieve_one(self, test_uuid):
    #     resp = self.session.get(self.spintop_api.get_link('tests.retrieve_one', test_uuid=test_uuid))
    #     test = resp.json()
    #     return get_json_serializer().deserialize(SpintopTestRecord, test)
        
    def _count(self, query):
        query_dict = query.as_dict()
        resp = self.session.get(self.spintop_api.get_link('tests.count'), params=query_dict)
        return test_count_schema.load(resp.json())['count']

    def _update(self, records, upsert=True):
        serialized = self._records_in_schema(records)
        return self.session.put(self.spintop_api.get_link('tests.update'), json=serialized)
    
    def _delete(self, query):
        query_dict = query.as_dict()
        return self.session.delete(self.spintop_api.get_link('tests.delete'), params=query_dict)



import socket
from contextlib import closing

from dateutil.tz import tzlocal

from collections.abc import Mapping
from contextlib import contextmanager
import cProfile, pstats, io

import yaml

try:
    import numpy as np
    isnan = np.isnan
except ImportError:
    isnan = lambda x: x is None


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]

def load_yaml_file(filename):
    with open(filename,'r') as readfile:
        data = readfile.read()
        return yaml.safe_load(data)
    
def write_yaml_file(filename, content):
    with open(filename,'w+') as writefile:
        yaml.dump(content, writefile)

def local_tz():
    return tzlocal()

@contextmanager
def profiling(output_filename):
    pr = cProfile.Profile()
    pr.enable()
    try:
        yield
    finally:
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        with open(output_filename, 'w+') as profile_stream:
            ps = pstats.Stats(pr, stream=profile_stream).sort_stats(sortby)
            ps.print_stats()



import os
from urllib.parse import urlparse, ParseResult

from sqlalchemy import create_engine, Table, MetaData, select, delete
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import NoSuchTableError, ProgrammingError
from sqlalchemy.sql import text

from .models import Base

def replace_database_in_uri(uri, database_name):
    parsed = urlparse(uri)
    with_database = ParseResult(
        scheme=parsed.scheme, 
        netloc=parsed.netloc, 
        path=database_name, 
        params=parsed.params, 
        query=parsed.query, 
        fragment=parsed.fragment
    )
    return with_database.geturl()

def engine_from_uri(uri, database_name=None):
    if database_name:
        uri = replace_database_in_uri(uri, database_name)
    return create_engine(uri)

class SQLOperations(object):
    def __init__(self, engine):
        self.engine = engine
        self.meta = MetaData(self.engine)
        self.Session = sessionmaker(engine)

    def new_session(self):
        return self.Session()

    def create_tables(self):
        Base.metadata.create_all(self.engine)

    def __getitem__(self, table_name):
        try:
            if not self.engine.dialect.has_table(self.engine, table_name):
                raise NoSuchTableError(table_name)
            
            return TableOperations(
                Table(table_name, self.meta, autoload=True, autoload_with=self.engine),
                self.engine
            )
        except (NoSuchTableError, ProgrammingError) as e:
            raise KeyError(str(e))

class TableOperations(object):
    def __init__(self, table, engine):
        self.table = table
        self.engine = engine
    
    def list_existing_indices(self, indices, column_name='index'):
        with self.engine.connect() as conn:
            rows = conn.execute(
                select(
                    [self.table.c[column_name]], 
                    self.table.c[column_name].in_(tuple(indices))
                )
            ).fetchall()

            return [row[0] for row in rows] # We requested only one field.

    def delete_indices(self, indices, column_name='index'):
        with self.engine.connect() as conn:
            conn.execute(
                delete(
                    self.table, 
                    self.table.c[column_name].in_(tuple(indices))
                )
            )

    def drop_table(self):
        self.table.drop()
        self.table.metadata.reflect()


import os
import sys
import click
import traceback
import logging
import threading
import time
from pprint import pformat, pprint
from tabulate import tabulate

from . import logs, storage
from .auth import AuthModule
from .compat import format_exception
from .errors import SpintopException
from .models import Query
from .env import Spintop


class CLIParseException(SpintopException):
    pass

logger = logs._logger('cli')

_spintop = None 
spintop_kwargs = {}
def spintop():
    global _spintop, spintop_kwargs
    if _spintop is None:
        _spintop = Spintop(**spintop_kwargs)
    return _spintop

@click.group()
@click.option('-v', '--verbose', is_flag=True, default=False)
@click.option('--url', default=None)
def _cli(verbose, url):
    global spintop_kwargs
    logs.VERBOSE = verbose
    spintop_kwargs.update(dict(verbose=verbose, api_url=url))

@_cli.command('site-data-dir')
def site_data_dir():
    click.echo(storage.SITE_DATA_DIR)
    
@_cli.command('login')
def login():
    spintop().login() # Check first before prompting

@_cli.command('login-cli')
@click.option('--username', default=None)
@click.option('--password', default=None)
def login_cli_command(username, password):
    if username is None and password is not None:
        raise CLIParseException('You cannot specify the password as option but not the username')
    
    spintop().assert_no_login() # Check first before prompting
    
    if username is None:
        username = click.prompt('Username')
        
    if password is None:
        password = click.prompt('Password', hide_input=True)
    
    spintop().login_user_pass(username, password)

@_cli.command('logout')
def logout_command():
    stored_username = spintop().stored_logged_username()
    if stored_username:
        if click.confirm('Logging out %s. Continue ?' % stored_username):
            spintop().logout()
    else:
        click.echo('No credentials stored.')

@_cli.command('userinfo')
def get_orgs():
    user_info = spintop().user_info
    click.echo(user_info)

@_cli.command('orginfo')
@click.argument('org_key')
def get_orgs(org_key):
    org_info = spintop().get_org_info(org_key)
    click.echo(org_info)

@_cli.command('ls-orgs')
def get_orgs():
    orgs = spintop().get_user_orgs()
    click.echo(orgs)

@_cli.command('ls-tests')
def list_tests():
    tests = spintop().tests.retrieve()
    tests = list(tests)
    for test in tests:
        click.echo(test.data.test_id)

@_cli.command('test-api-private')
def test_private_command():
    print(spintop().test_private_endpoint().json())
    
@_cli.command('force-corrupt-access-token')
def force_corrupt_access_token():
    spintop_api = spintop()
    spintop_api.auth.credentials.access_token = spintop_api.auth.credentials.access_token[0:-4]
    spintop_api.auth.save_credentials()


def cli(args=None, internal=True):
    try:
        return _cli(args, prog_name='spintop')
    except Exception as e:
        if logs.VERBOSE:
            click.echo(traceback.format_exc())
        else:
            click.echo(format_exception(e))
    except SystemExit:
        if not internal:
            raise

def main():
    cli(internal=False)

if __name__ == '__main__':
    main()


import dateutil.parser
from typing import Union
from functools import lru_cache
from datetime import datetime
from time import mktime
from copy import deepcopy, copy

from dataclasses_serialization.serializer_base import (
    dataclass,
    Serializer, 
    dict_to_dataclass, 
    dict_serialization, 
    noop_serialization,
    dict_deserialization,
    list_deserialization, 
    union_deserialization,
    noop_deserialization, 
)

from .base import BaseDataClass
from .test_records import SpintopTestRecord, SpintopSerializedFlatTestRecord

from spintop.utils import local_tz, isnan


def get_serializer():
    return get_bson_serializer()

@lru_cache(maxsize=128)
def get_bson_serializer():
    """Make or return the singleton SpintopBSONSerializer"""
    from dataclasses_serialization.bson import BSONSerializer
    serializer = _SpintopSerializer(BSONSerializer)
    return serializer

@lru_cache(maxsize=128)
def get_json_serializer(datetime_serializer=lambda datetime_obj: datetime_obj.isoformat()):
    from dataclasses_serialization.json import JSONSerializer
    serializer = _SpintopSerializer(JSONSerializer, datetime_serializer=datetime_serializer)
    return serializer

def _copy_serializer(serializer):
    return 

class _SpintopSerializer(Serializer):
    def __init__(self, source_ser, datetime_serializer=noop_serialization):
        super().__init__(
            serialization_functions=copy(source_ser.serialization_functions),
            deserialization_functions=copy(source_ser.deserialization_functions)
        )
        self._make_serializer(datetime_serializer=datetime_serializer)

    def __post_init__(self):
        for ser_map in (self.serialization_functions, self.deserialization_functions):
            try:
                del ser_map[dataclass]
            except KeyError:
                pass
        self.deserialization_functions.setdefault(Union, union_deserialization(deserialization_func=self.deserialize))

    def _make_serializer(self, datetime_serializer=lambda datetime_obj: datetime_obj):
        
        # The base serializer has an entry of type: (str, int, ...): noop_serializer
        # Because the key contains many types, if we want to overwrite one of them, we
        # get the risk that the default one gets picked. Therefore, we split the key
        # in individual functions before adding our customs one which may overwrite these
        # ones.
        _split_key_of_op(self.deserialization_functions, noop_deserialization)
        _split_key_of_op(self.serialization_functions, noop_serialization)

        self.register(
            dict,
            serialization_func = lambda dct: dict_serialization(dct, key_serialization_func=self.serialize, value_serialization_func=self.serialize),
            deserialization_func = lambda cls, dct: dict_deserialization(cls, dct, key_deserialization_func=self.deserialize, value_deserialization_func=self.deserialize)
        )

        self.register(
            list,
            serialization_func = lambda lst: list(map(self.serialize, lst)),
            deserialization_func = lambda cls, lst: list_deserialization(cls, lst, deserialization_func=self.deserialize)
        )

        @self.register_serializer(float)
        def serialize_nan(f_value):
            if isnan(f_value):
                return None
            else:
                return f_value

        @self.register_serializer(datetime)
        def serialize_datetime(obj):
            # obj = obj.astimezone(local_tz())
            return datetime_serializer(obj)

        @self.register_deserializer(datetime)
        def deserialize_datetime(cls, obj):
            datetime_obj = None
            if isinstance(obj, str):
                datetime_obj = dateutil.parser.parse(obj)
            elif isinstance(obj, datetime):
                datetime_obj = obj 
            else:
                datetime_obj = datetime.fromtimestamp(obj)
            return datetime_obj
            # return datetime_obj.astimezone(local_tz())

        @self.register_serializer(BaseDataClass)
        def serialize(obj):
            """Custom serializer for BaseDataClass that extracts the static
            attribute '_type' into the serialized dict for future deserialization."""
            data = dict_serialization(dict(obj.__dict__), key_serialization_func=self.serialize, value_serialization_func=self.serialize)
            data.update({'_type': obj._type})
            return data

        @self.register_deserializer(BaseDataClass)
        def deserialize(cls, obj):
            """ Custom deserializer that infers the BaseDataClass subclass based
            on the _type attribute. 
            """
            # Strips the _type field and returns the sub cls, if found. Else returns the cls
            # as passed in parameter
            cls, obj = BaseDataClass.cls_data_from_dict(cls, obj)
            return dict_to_dataclass(cls, obj, deserialization_func=self.deserialize)

        
        @self.register_serializer((SpintopTestRecord, SpintopSerializedFlatTestRecord))
        def serialize_spintop_test_record(obj):
            return self.serialize(obj.as_dict())

        @self.register_deserializer(SpintopTestRecord)
        def deserialize_spintop_test_record(cls, obj):
            result = {}
            if isinstance(obj, SpintopSerializedFlatTestRecord):
                result = obj.as_dict()
            else:
                if 'test_record' in obj:
                    result['test_record'] = self.deserialize(BaseDataClass, obj['test_record'])
                
                if 'features' in obj:
                    result['features'] = [self.deserialize(BaseDataClass, feat) for feat in obj['features']]

            return cls.from_dict(result)
        

def _split_key_of_op(op_map, op_fn):
    key_is = None

    for key, fn in op_map.items():
        if fn is op_fn:
            key_is = key
            break
    
    if key_is is None:
        raise ValueError(f'Did not find {op_fn} in map.')
    elif not isinstance(key_is, tuple):
        return
    else:
        del op_map[key_is]
    
        for key in  key_is:
            op_map[key] = op_fn




import uuid
import requests
import platform

from spintop import __version__

def spintop_session(user_agent=f'SPINTOP/{__version__} HOST/{platform.node()} MACHINEID/{uuid.getnode()}'):
    session = requests.Session()
    session.headers.update({'User-Agent': user_agent})
    return session


import re
from functools import wraps
from collections import OrderedDict, defaultdict
from .test_records import FeatureRecord, TestRecordSummary
from .serialization import get_json_serializer


from spintop.utils.dict_ops import (
    update,
    deepen_dict,
    flatten_dict
)

def record_call(exclude_top_level=False):
    def _record_call(fn):
        @wraps(fn)
        def _recorded(self, arg):
            self._record_call(fn.__name__, arg, exclude_top_level)
            return fn(self, arg)
        return _recorded
    return _record_call

class Query():
    model_type = FeatureRecord

    def __init__(self):
        # The field key must equal exactly value, or re.search if a compiled regex
        self._value_equals = dict()

        # The field of type list named key must contain value
        self._list_contains = dict()

        # The field named key must equal one of the sub value in the list value
        self._value_equals_one_of = dict()

        self._calls = OrderedDict()
        self._top_level_excluded_calls = set()

    @property
    def value_equals(self):
        return self._value_equals

    @property
    def list_contains(self):
        return self._list_contains

    @property
    def value_equals_one_of(self):
        return self._value_equals_one_of

    def _record_call(self, fn_name, arg, exclude_top_level=False):
        self._calls[fn_name] = arg
        if exclude_top_level:
            self._top_level_excluded_calls.add(fn_name)

    def add_call(self, fn_name, arg):
        getattr(self, fn_name)(arg)

    @record_call(exclude_top_level=True)
    def name_regex_is(self, regex):
        self._value_equals[self.model_type.name.name_] = re.compile(regex)
        return self
    
    def type_is(self, cls):
        return self.type_is_str(cls._type)

    @record_call(exclude_top_level=True)
    def type_is_str(self, cls_str):
        self._value_equals['_type'] = cls_str
        return self

    def type_any_of(self, classes):
        return self.type_any_of_str([cls._type for cls in classes])
    
    @record_call(exclude_top_level=True)
    def type_any_of_str(self, classes_str):
        self._value_equals_one_of['_type'] = classes_str
        return self

    def tag_is(self, **tag_values):
        return self.tag_is_raw(tag_values)

    def tag_regex_is(self, **tag_regexes):
        return self.tag_is_raw({key: re.compile(value) for key, value in tag_regexes.items()})

    @record_call(exclude_top_level=False)
    def tag_is_raw(self, tag_values):
        base_name = self.model_type.test_id.tags.name_
        for key, value in tag_values.items():
            self._value_equals['{}.{}'.format(base_name, key)] = value
        return self

    @record_call(exclude_top_level=False)
    def test_uuid_is(self, test_uuid):
        self._value_equals[self.model_type.test_id.test_uuid.name_] = test_uuid
        return self
    
    @record_call(exclude_top_level=False)
    def test_uuid_any_of(self, test_uuids):
        self._value_equals_one_of[self.model_type.test_id.test_uuid.name_] = test_uuids
        return self
    
    @record_call(exclude_top_level=False)
    def testbench_name_is(self, testbench_name):
        self._value_equals[self.model_type.test_id.testbench_name.name_] = testbench_name
        return self
    
    def outcome_is(self, **outcome_attributes):
        return self.outcome_is_raw(outcome_attributes)

    @record_call(exclude_top_level=False)
    def outcome_is_raw(self, outcome_attributes):
        for field_name, value in outcome_attributes.items():
            field = getattr(self.model_type.outcome, field_name)
            self._value_equals[field.name_] = value
        return self
    
    def dut_match(self, id=None, version=None):
        return self.dut_match_raw({'id': id, 'version': version})

    @record_call(exclude_top_level=False)
    def dut_match_raw(self, id_and_version):
        _id = id_and_version.get('id', None)
        _version = id_and_version.get('version', None)
        if _id is not None:
            self._value_equals[self.model_type.test_id.dut.id.name_] = re.compile(_id)
        if _version is not None:
            self._value_equals[self.model_type.test_id.dut.version.name_] = re.compile(_version)
        return self

    def custom_match(self, match):
        self._value_equals.update(match)
        return self

    def __eq__(self, other_q):
        same_len = len(self._calls) == len(other_q._calls)
        if not same_len:
            return False

        for key, value in self._calls.items():
            if key not in other_q._calls or other_q._calls[key] != value:
                return False
        
        return True

    def __repr__(self):
        return '{}(eq={}, one-of={}, contains={})'.format(
            self.__class__.__name__,
            repr(self._value_equals),
            repr(self._value_equals_one_of),
            repr(self._list_contains)
        )

    KEY_SEPARATOR = '::'
    def as_dict(self):
        serializer = get_json_serializer()
        deep_dict = serializer.serialize(self._calls)
        flat_dict = flatten_dict(deep_dict)
        return {self.KEY_SEPARATOR.join(key) : value for key, value in flat_dict.items()}

    @classmethod
    def from_dict(cls, _flat_separated_calls):
        calls = cls.parse_query_dict(_flat_separated_calls)
        query = cls()
        for fn_name, arg in calls.items():
            query.add_call(fn_name, arg)
        return query

    @classmethod
    def parse_query_dict(cls, _flat_separated_calls):
        flat_dict = {tuple(key.split(cls.KEY_SEPARATOR)): value for key, value in _flat_separated_calls.items()}
        return deepen_dict(flat_dict)

    def copy(self):
        return self.__class__.from_dict(self.as_dict())

    def create_top_level_query(self):
        data = self.as_dict()
        for key in self._top_level_excluded_calls:
            del data[key]
        
        top_level_query = self.from_dict(data)
        # Filter for top level
        return top_level_query.type_is(TestRecordSummary).name_regex_is('')

def multi_query_serialize(**queries):
    """ Serialized as key_subkey = subvalue for key key, query in queries.
    """

    result = {}

    for key, query in queries.items():
        if '_' in key:
            raise ValueError('Query key {!r} cannot contain an underscore.'.format(key))
        if query:
            for subkey, subvalue in query.as_dict().items():
                result['{}_{}'.format(key, subkey)] = subvalue
    
    return result

def multi_query_deserialize(_dict):
    queries = {}

    for key, value in _dict.items():
        query_name, fn_name = split_key(key)
        update(queries, {
            query_name: Query.parse_query_dict({fn_name: value})
        })
    
    return {query_name: Query.from_dict(query_dict) for query_name, query_dict in queries.items()}

def split_key(key):
    split_key = key.split('_')
    return split_key[0], '_'.join(split_key[1:])


from collections.abc import Mapping
from itertools import chain, starmap

class DefaultValue(object):
    def __init__(self, value_or_fn):
        self.value_or_fn = value_or_fn

    def resolve(self):
        if callable(self.value_or_fn):
            return self.value_or_fn()
        else:
            return self.value_or_fn

DEFAULT_PLACEHOLDER = DefaultValue(None)

def update(target, updater):
    """Deep update of dict target using updater. 
    
    The special DEFAULT_PLACEHOLDER object allows to set default values"""
    for key, value in updater.items():
        if isinstance(value, Mapping):
            target[key] = update(target.get(key, {}), value)
        else:
            if is_default(value) and key in target:
                # Do no replace existing value with the default placeholder
                pass
            else:
                target[key] = value
    return target

def is_default(value):
    return isinstance(value, DefaultValue)

def replace_defaults(target):
    for key, value in target.items():
        if isinstance(value, Mapping):
            replace_defaults(value)
        else:
            if is_default(value):
                target[key] = value.resolve()
    return target

def deepen_value_in_dict(key_prefix, value):
    result = {}
    next_dict = result
    for key in key_prefix[:-1]:
        next_dict[key] = {}
        next_dict = next_dict[key]
    next_dict[key_prefix[-1]] = value
    
    return result

def deepen_dict(flat_dict):
    result = {}
    for key, value in flat_dict.items():
        update(result, deepen_value_in_dict(key, value))
    return result

def flatten_dict(dictionary):
    """Flatten a nested dictionary structure"""

    tuplify = lambda key: key if isinstance(key, tuple) else (key,)

    def unpack(parent_key, parent_value):
        """Unpack one level of nesting in a dictionary"""
        try:
            items = parent_value.items()
        except AttributeError:
            # parent_value was not a dict, no need to flatten
            yield (parent_key, parent_value)
        else:
            for key, value in items:
                yield (parent_key + tuplify(key), value)


    # Put each key into a tuple to initiate building a tuple of subkeys
    dictionary = {tuplify(key): value for key, value in dictionary.items()}

    while True:
        # Keep unpacking the dictionary until all value's are not dictionary's
        dictionary = dict(chain.from_iterable(starmap(unpack, dictionary.items())))
        if not any(isinstance(value, dict) for value in dictionary.values()):
            break

    return dictionary


import os

import requests 
import platform
import functools

from contextlib import contextmanager
from pprint import pformat
from uuid import uuid4

from functools import partial
from ..errors import SpintopException

from ..auth import AuthModule
from ..logs import _logger
from ..compat import VERSION
from ..storage import TEMP_DIR

from .tests_facade import SpintopAPIPersistenceFacade
from .analytics import SpintopAPIAnalytics

from spintop.utils.requests_utils import spintop_session

logger = _logger('spintop-api')

TEMPORARY_DOWNLOAD_FOLDER = os.path.join(TEMP_DIR, '.spintop-api-download-cache')
if not os.path.exists(TEMPORARY_DOWNLOAD_FOLDER):
    try:
        os.makedirs(TEMPORARY_DOWNLOAD_FOLDER)
    except:
        logger.warning(f'Unable to makedirs({TEMPORARY_DOWNLOAD_FOLDER}) even though file does not exist.')
    
class SpintopClientError(SpintopException):
    pass

class SpintopAPIError(SpintopException):
    def __init__(self, resp):
        try:
            json = resp.json()
        except Exception:
            json = {}
        
        error_type = json.get('error_type', 'UNKNOWN')
        error_messages = json.get('error_messages', {})
        messages_as_string = '\n\t'.join(['%s: %s' % (key, value) for key, value in error_messages.items()])
        
        if not messages_as_string:
            messages_as_string = json.get('error_message') # singular, the direct exception message
            
        message = f'Request {resp.request.method} {resp.url} A {resp.status_code} error of type "{error_type}" occured while contacting API.\n\t{messages_as_string}'
        
        super().__init__(message)

        self.error_type = error_type
        self.resp = resp

def refresh_if_fail_gen(auth_module):
    try:
        yield
    except SpintopAPIError as e:
        if e.resp.status_code == 401 and e.error_type in ('auth_token_verification_failed',) :
            # Attempt ONCE to refresh the credentials and retry.
            auth_module.refresh_credentials()
            yield
        else:
            raise

def SpintopAPISession(prefix, auth_module):
    if prefix is None:                                                                                                                                                                                                                                                                                                                             
        prefix = ""                                                                                                                                                                                                                                                                                                                                
    else:                                                                                                                                                                                                                                                                                                                                          
        prefix = prefix.rstrip('/')                                                                                                                                                                                                                                                                                                      

    def attempt_request(f, method, url, *args, **kwargs):
        if 'headers' not in kwargs: kwargs['headers'] = {}
        kwargs['headers'].update(auth_module.get_auth_headers())
        resp = f(method, prefix + url, *args, **kwargs)           
        if not resp.ok:
            raise SpintopAPIError(resp)
        return resp
    
    def new_request(prefix, *args, **kwargs):
        
        try:
            resp = attempt_request(*args, **kwargs)
        except SpintopAPIError as e:
            if e.resp.status_code == 401 and e.error_type in ('auth_token_verification_failed', 'auth_missing_header') :
                # Attempt ONCE to refresh the credentials and retry.
                auth_module.refresh_credentials()
                resp = attempt_request(*args, **kwargs)
            else:
                raise
        return resp
    
    s = spintop_session()
    s.request = partial(new_request, prefix, s.request)
    return s

class SpintopAPIClientModule(object):
    def __init__(self, api_url, selected_org_id=None, auth=None):
        self._session = None

        self.auth = auth
        
        self._selected_org_id = selected_org_id
        self._tests_facade = SpintopAPIPersistenceFacade(self)
        self._analytics = SpintopAPIAnalytics(self)
        
        self._api_url = api_url
        self._user_info = None
        self._api_spec = None
        self._default_org = None
        self._machine = None

    @property
    def api_url(self):
        if callable(self._api_url):
            return self._api_url()
        else:
            return self._api_url

    @api_url.setter
    def api_url(self, value):
        self._session = None
        self._api_url = value

    @property
    def session(self):
        if not self._session:
            self._session = self.new_session()
        return self._session

    @property
    def user_info(self):
        if self._user_info is None:
            self._user_info = self.session.get(self.get_link('tenancy.user_info')).json()
        return self._user_info

    @property
    def api_spec(self):
        if self._api_spec is None:
            # Use a new session here, since we need api_spec to init auth during session retrival.
            self._api_spec = self.new_session().get('/api/').json()
        return self._api_spec
    
    @property
    def selected_org_id(self):
        if self._selected_org_id is None:
            possible_orgs = [org['key'] for org in self.user_info['organizations']]
            selected_org = possible_orgs[0]
            logger.info('No org specified, using the first defined by the API for this user: %s' % selected_org)
            self._selected_org_id = selected_org
        
        return self._selected_org_id

    @property
    def tests(self):
        return self._tests_facade

    @property
    def analytics(self):
        return self._analytics

    def login(self):
        """Interactive login"""
        self.auth.assert_no_login() # Check first before prompting

        with self.login_browser() as device_flow_response:
            logger.info(f'User code is {device_flow_response.user_code!r}')
            logger.info(f'If this device is headless (without a browser), you may visit this link to authorize this device on another machine: {device_flow_response.verification_uri_complete}')

    def login_user_pass(self, username, password):
        self.auth.login_user_pass(username, password)

    def login_browser(self):
        return self.auth.login_device_flow()

    def stored_logged_username(self):
        if self.auth.credentials:
            return self.auth.credentials.get('username')
        else:
            return None
    
    def logout(self):
        return self.auth.logout()
    
    def get_user_orgs(self):
        return self.auth.user_orgs

    def new_session(self):
        return SpintopAPISession(self.api_url, self.auth)
    
    def reset_session(self):
        self._session = None

    def test_private_endpoint(self):
        resp = self.session.get(self.get_link('core.private'))
        return resp

    def get_org_info(self, org_id=None):
        return self.session.get(self.get_link('tenancy.org_info', org_id=org_id)).json()

    def create_org(self, org_id):
        return self.session.post(self.get_link('tenancy.create_org', org_id=org_id)).json()

    def get_tests_env(self, org_id=None):
        """ An env that points to the API itself """
        return self.session.get(self.get_link('tests.env', org_id=org_id)).json()

    def get_tests_env_direct_database_context(self, org_id=None):
        """ An env that points to the API's underlying database (with the API's URI)"""
        if org_id is None: org_id = self.selected_org_id
        env = self.get_tests_env(org_id)
        org_info = self.get_org_info(org_id)
        return env.new_database_context(org_info['database_name'])

    def get_link(self, link_name, org_id=None, **kwargs):
        links = self.api_spec['_links']
        base_link = links[link_name]

        if '<org_id>' in base_link:
            if org_id is None:
                org_id = self.selected_org_id

            kwargs['org_id'] = org_id

        link = base_link
        for key, value in kwargs.items():
            link = link.replace(f'<{key}>', value)

        if '<' in link or '>' in link:
            raise SpintopClientError(f'Link {base_link!r} is missing placeholder values. Received {kwargs.keys()!r}')

        return link
        
        


from .test_records import *

from .base import (
    type_of,
    serialized_type_of,
    is_type_of,
    type_dict_of,
    is_serialized_type_of
)

from .queries import Query, multi_query_deserialize, multi_query_serialize

from .serialization import get_serializer, get_bson_serializer, get_json_serializer

from .view import DataClassPrimitiveView, ComplexPrimitiveView, update

from .view_helpers import (
    value_of_feature_named
)



import functools

from .base import SpintopAPIClientModule

from ..services.auth0 import Auth0Provider


class NoAuthModule(object):
    def get_auth_headers(self):
        return {}

    def refresh_credentials(self):
        raise NotImplementedError('Cannot refresh credentials without an auth module.')


class SpintopAPIAuthProvider(Auth0Provider):
    def __init__(self, *auth0_args, payload_claims={}, **auth0_kwargs):
        super().__init__(*auth0_args, **auth0_kwargs)
        self.decoder = SpintopAccessTokenDecoder(payload_claims)

    @functools.lru_cache(maxsize=128)
    def get_user_orgs(self, access_token):
        token_payload = self.jwt_verify(access_token)
        return self.decoder.get_user_orgs(token_payload)

class SpintopAccessTokenDecoder(object):
    def __init__(self, payload_claims):
        self.payload_claims = payload_claims
    
    def _claim_name(self, key):
        return self.payload_claims.get(key)

    def get_user_orgs(self, token_payload):
        groups_from_auth = token_payload.get(self._claim_name('authorization'), {}).get('groups', [])
        scopes = token_payload.get('scope').split(' ')
        groups_from_scope = [scope.strip('group:') for scope in scopes if scope.startswith('group:')]
        return groups_from_auth + groups_from_scope


class SpintopAPISpecAuthBootstrap(SpintopAPIClientModule):
    """A no auth client module to bootstrap the retrieval of api spec before the auth_provider exists,
    since its parameters are returned by the /api endpoint itself."""
    def __init__(self, api_url):
        super().__init__(api_url, auth=NoAuthModule())
        self._api_spec = None
        self._auth_provider = None

    def get_auth_spec(self):
        api_spec = self.api_spec
        auth_spec = api_spec['auth']
        return dict(
            enabled=auth_spec.get('enabled'),
            domain=auth_spec.get('domain'),
            client_id=auth_spec.get('machine_client_id'),
            audience=auth_spec.get('audience'),
            jwks_url=auth_spec.get('jwks_url'),
            user_info_url=auth_spec.get('user_info_url'),
            payload_claims=auth_spec.get('payload_claims', {})
        )

    def get_auth_provider(self):
        if self._auth_provider is None:
            auth_spec = self.get_auth_spec()
            if auth_spec.pop('enabled'):
                try:
                    self._auth_provider = SpintopAPIAuthProvider(**auth_spec)
                except ValueError:
                    raise ValueError(f'API Spec is invalid: unable to create auth0 provider. {auth_spec}')
        return self._auth_provider

class BackendSpintopApiAuthProvider(SpintopAPIAuthProvider):
    def refresh_access_token(self, refresh_token=None):
        # Use client credentials grant instead of refreshing token.
        # This requires a client_id and secret_key.
        return self.client_credentials()

class SpintopBackendAPISpecAuthBootstrap(SpintopAPISpecAuthBootstrap):
    def __init__(self, api_url, client_id, secret_key):
        super().__init__(api_url)

        self.client_id = client_id
        self.secret_key = secret_key

    def get_auth_provider(self):
        if self._auth_provider is None:
            auth_spec = self.get_auth_spec()
            auth_spec['client_id'] = self.client_id
            auth_spec['secret_key'] = self.secret_key
            if auth_spec.pop('enabled'):
                try:
                    self._auth_provider = BackendSpintopApiAuthProvider(**auth_spec)
                except ValueError:
                    raise ValueError(f'API Spec is invalid: unable to create auth0 provider. {auth_spec}')
        return self._auth_provider
    

def create_backend_auth_bootstrap_factory(client_id, secret_key):
    """Create a factory that uses the machine to machine auth flow using the client_id and secret_key."""
    def auth_bootstrap_factory(api_url):
        return SpintopBackendAPISpecAuthBootstrap(api_url, client_id, secret_key)
    
    return auth_bootstrap_factory



from dataclasses import dataclass, fields, field, Field

class MetaBaseDataClass(type):
    def __getattr__(cls, key):
        """ Allows fields to be accessed from class. I.e. 
        
        @dataclass()
        class MyDataClass(BaseDataClass):
            my_field: str
            
        MyDataClass.my_field.name == 'my_field'
        """
        try:
            return cls.get_field(key)
        except AttributeError:
            return super().__getattribute__(key)
    
    def get_field(cls, key, parent_fields=()):
        # no cache here has the cls might be being build.
        for field in fields(cls):
            if field.name == key:
                return MetaField(cls, field, parents=parent_fields)
        else:
            raise AttributeError(key)


class MetaField(Field):
    def __init__(self, cls, field, parents=()):
        self.__type = cls
        self.__parents = parents
        self.__field = field
    
    @property
    def name_(self):
        return '.'.join(self.name_tuple_)

    @property
    def type_(self):
        return self.__field.type

    @property
    def container_type_(self):
        if self.__parents:
            return self.__parents[0].__type
        else:
            return self.__type

    @property
    def name_tuple_(self):
        return tuple(parent.__field.name for parent in (self.__parents + (self,))) 

    def value_of_serialized(self, serialized_obj):
        for field_name in self.name_tuple_:
            serialized_obj = serialized_obj[field_name]
        
        return serialized_obj

    def is_container_of_cls(self, cls):
        return cls is self.container_type_ or issubclass(cls, self.container_type_)

    def __getattr__(self, key):
        try:
            return self.__field.type.get_field(key, parent_fields=self.__parents + (self,))
        except AttributeError:
            return getattr(self.__field, key)

    def __hash__(self):
        return hash(self.__field)
    
    def __eq__(self, other):
        return self.__field == other.__field

    
# def make_operation(field, op_on_value):
#     return lambda real_obj: op_on_value(getattr(real_obj, field.name))



import importlib

from abc import ABC, abstractmethod

from spintop.models import SpintopTestRecordCollection
from spintop.transforms import Transformer

class Generator(Transformer):
    """Transformers natively support a generator behaviour through the __call__ function."""
    
class GeneratorFromModule(Generator):
    def __init__(self, module):
        super(GeneratorFromModule, self).__init__()
        self.module = module
        
    @classmethod
    def from_module_name(cls, module_name):
        return cls(
            importlib.import_module(module_name)
        )
        
    def __call__(self, *args, **kwargs):
        """Override."""
        return self.module.generate(*args, **kwargs)

def generator_from_module_or_module_name(module_or_module_name):
    if isinstance(module_or_module_name, str):
        generator_obj = importlib.import_module(module_or_module_name)
    else:
        generator_obj = module_or_module_name
    return GeneratorFromModule(generator_obj)